{
  "cells": [
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "# 연습장\n",
        "\n",
        "JiyunLim  \n",
        "2023-08-28"
      ],
      "id": "c4d39899-bafb-42aa-b3ff-8a057177da2c"
    },
    {
      "cell_type": "code",
      "execution_count": 1,
      "metadata": {},
      "outputs": [],
      "source": [
        "try:\n",
        "    from tqdm import tqdm\n",
        "except ImportError:\n",
        "    def tqdm(iterable):\n",
        "        return iterable"
      ],
      "id": "56bc1bd8-4feb-4e5d-acd9-9c6ea90ac40c"
    },
    {
      "cell_type": "code",
      "execution_count": 2,
      "metadata": {},
      "outputs": [],
      "source": [
        "import torch\n",
        "import torch.nn.functional as F\n",
        "from torch_geometric_temporal.nn.recurrent import AGCRN\n",
        "\n",
        "from torch_geometric_temporal.dataset import ChickenpoxDatasetLoader\n",
        "from torch_geometric_temporal.signal import temporal_signal_split"
      ],
      "id": "cd83d072-ebeb-4f6d-a477-27f9931cb583"
    },
    {
      "cell_type": "code",
      "execution_count": 3,
      "metadata": {},
      "outputs": [],
      "source": [
        "loader = ChickenpoxDatasetLoader()\n",
        "\n",
        "dataset = loader.get_dataset(lags=8)\n",
        "\n",
        "train_dataset, test_dataset = temporal_signal_split(dataset, train_ratio=0.2)"
      ],
      "id": "abaeb05d-3889-4f71-a599-2691cae57c03"
    },
    {
      "cell_type": "code",
      "execution_count": 15,
      "metadata": {},
      "outputs": [],
      "source": [
        "class RecurrentGCN(torch.nn.Module):\n",
        "    def __init__(self, node_features):\n",
        "        super(RecurrentGCN, self).__init__()\n",
        "        self.recurrent = AGCRN(number_of_nodes = 20,\n",
        "                              in_channels = node_features, # 노드특징 수\n",
        "                              out_channels = 16,  # 필터 수\n",
        "                              K = 2,\n",
        "                              embedding_dimensions = 4)\n",
        "        self.linear = torch.nn.Linear(16, 1)\n",
        "\n",
        "    def forward(self, x, e, h):\n",
        "        h_0 = self.recurrent(x, e, h)\n",
        "        y = F.relu(h_0)\n",
        "        y = self.linear(y)\n",
        "        return y, h_0"
      ],
      "id": "3ff50386-7e7b-4883-9995-16287909693b"
    },
    {
      "cell_type": "code",
      "execution_count": 16,
      "metadata": {},
      "outputs": [],
      "source": [
        "model = RecurrentGCN(node_features = 8)\n",
        "model"
      ],
      "id": "91e65d6b-9b71-42f2-9098-19dfea0c39df"
    },
    {
      "cell_type": "code",
      "execution_count": 17,
      "metadata": {},
      "outputs": [
        {
          "output_type": "stream",
          "name": "stderr",
          "text": [
            "100%|██████████| 30/30 [00:58<00:00,  1.96s/it]"
          ]
        }
      ],
      "source": [
        "model = RecurrentGCN(node_features = 8)\n",
        "\n",
        "optimizer = torch.optim.Adam(model.parameters(), lr=0.01)\n",
        "\n",
        "model.train()\n",
        "\n",
        "e = torch.empty(20, 4) # 노드수 x 임베딩 차원 수\n",
        "\n",
        "torch.nn.init.xavier_uniform_(e) # 임베딩 벡터 초기화\n",
        "\n",
        "for epoch in tqdm(range(30)):\n",
        "    cost = 0\n",
        "    h = None\n",
        "    for time, snapshot in enumerate(train_dataset):\n",
        "        x = snapshot.x.view(1, 20, 8)\n",
        "        y_hat, h = model(x, e, h)\n",
        "        cost = cost + torch.mean((y_hat-snapshot.y)**2)\n",
        "    cost = cost / (time+1)\n",
        "    cost.backward()\n",
        "    optimizer.step()\n",
        "    optimizer.zero_grad()"
      ],
      "id": "c31247fe-593d-4fd0-8be7-160c34686057"
    },
    {
      "cell_type": "code",
      "execution_count": 18,
      "metadata": {},
      "outputs": [
        {
          "output_type": "stream",
          "name": "stdout",
          "text": [
            "MSE: 1.0056"
          ]
        }
      ],
      "source": [
        "model.eval()\n",
        "cost = 0\n",
        "for time, snapshot in enumerate(test_dataset):\n",
        "    x = snapshot.x.view(1, 20, 8)\n",
        "    y_hat, h = model(x, e, h)\n",
        "    cost = cost + torch.mean((y_hat-snapshot.y)**2)\n",
        "cost = cost / (time+1)\n",
        "cost = cost.item()\n",
        "print(\"MSE: {:.4f}\".format(cost))"
      ],
      "id": "b4547484-d35e-4026-ad9c-d3e647386086"
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "AGCRN 모델은 그래프 데이터에서 노드 간의 동적 상호 작용을 학습하여\n",
        "시계열 데이터의 패턴을 모델링한다. 이를 통해 시간적 상관 관계와 노드\n",
        "간의 영향력을 효과적으로 표현한다. 모델 내부의 가중치와 임베딩 벡터들은\n",
        "시간 스텝별로 입력 데이터와 상호 작용하면서 데이터의 변화와 관계를\n",
        "이해하는 데 사용된다. 따라서 AGCRN 모델은 시계열 데이터의 특징을\n",
        "추출하고 상관 관계를 모델링하는데 필터와 유사한 역할을 수행한다.\n",
        "\n",
        "요약하면, self.recurrent에 해당하는 AGCRN 모델은 그래프 내에서 노드 간의\n",
        "상호 작용을 표현하고 시계열 데이터의 패턴과 상관 관계를 추출하는데\n",
        "필터와 유사한 역할을 한다.\n",
        "\n",
        "------------------------------------------------------------------------"
      ],
      "id": "2704e7a0-5ba3-45c6-9c2b-eda4c1d46262"
    },
    {
      "cell_type": "code",
      "execution_count": 19,
      "metadata": {},
      "outputs": [],
      "source": [
        "# modules \n",
        "import numpy as np\n",
        "import pandas as pd\n",
        "import matplotlib.pyplot as plt\n",
        "\n",
        "# pytorch geometric temporal\n",
        "import torch_geometric_temporal\n",
        "from torch_geometric_temporal.signal import temporal_signal_split\n",
        "from torch_geometric_temporal.signal.static_graph_temporal_signal import StaticGraphTemporalSignal\n",
        "from torch_geometric_temporal.nn.recurrent import AGCRN\n",
        "\n",
        "# torch\n",
        "import torch\n",
        "import torch.nn.functional as F\n",
        "\n",
        "\n",
        "# read data\n",
        "def makedict(FX,W=None,node_ids=None):\n",
        "    T,N = np.array(FX).shape\n",
        "    if W==None:\n",
        "        W = pd.DataFrame(FX).corr().to_numpy().reshape(-1).tolist() # corr 로 weights\n",
        "    if node_ids==None:\n",
        "        node_ids = ['node'+str(n) for n in range(N)]\n",
        "    _dict={\n",
        "        'edges':[[i,j] for i in range(N) for j in range(N)], \n",
        "        'node_ids': node_ids,\n",
        "        'weights':W,\n",
        "        'FX':FX\n",
        "    }\n",
        "    return _dict\n",
        "\n",
        "class Loader(object):\n",
        "    def __init__(self, data_dict):\n",
        "        self._dataset = data_dict\n",
        "    \n",
        "    def _get_edges(self):\n",
        "        self._edges = np.array(self._dataset[\"edges\"]).T\n",
        "\n",
        "    def _get_edge_weights(self):\n",
        "        edge_weights = np.array(self._dataset[\"weights\"]).T\n",
        "        self._edge_weights = edge_weights\n",
        "\n",
        "    def _get_targets_and_features(self):\n",
        "        stacked_target = np.stack(self._dataset[\"FX\"])\n",
        "        self.features = np.stack([\n",
        "            stacked_target[i : i + self.lags, :].T\n",
        "            for i in range(stacked_target.shape[0] - self.lags)\n",
        "        ])\n",
        "        self.targets = np.stack([\n",
        "            stacked_target[i + self.lags, :].T\n",
        "            for i in range(stacked_target.shape[0] - self.lags)\n",
        "        ])\n",
        "\n",
        "\n",
        "    def get_dataset(self, lags: int = 4) -> StaticGraphTemporalSignal:\n",
        "        self.lags = lags\n",
        "        self._get_edges()\n",
        "        self._get_edge_weights()\n",
        "        self._get_targets_and_features()\n",
        "        dataset = StaticGraphTemporalSignal(\n",
        "            self._edges, self._edge_weights, self.features, self.targets\n",
        "        )\n",
        "        dataset.node_ids = self._dataset['node_ids']\n",
        "        return dataset\n",
        "    \n",
        "\n",
        "# device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')\n",
        "device = torch.device('cpu')"
      ],
      "id": "2a99f2a4-1ff0-4fd0-a6db-ab3d03552f94"
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "`-` solar data"
      ],
      "id": "5f08a2d4-7ff6-4c99-b680-0e8f97b56307"
    },
    {
      "cell_type": "code",
      "execution_count": 20,
      "metadata": {},
      "outputs": [],
      "source": [
        "# read dataframe \n",
        "df = pd.read_csv('../SOLAR/data_eng_230710.csv')\n",
        "\n",
        "# make y, y_upper, y_period, time, regions \n",
        "y = df.loc[:,'Bukchoncheon':'Gyeongju-si'].to_numpy()\n",
        "yU = df.loc[:,'Bukchoncheon_Upper':'Gyeongju-si_Upper'].to_numpy()\n",
        "yP = np.divide(y, yU+1e-10)\n",
        "\n",
        "t = df.loc[:,'date']\n",
        "regions = list(df.loc[:, 'Bukchoncheon':'Gyeongju-si'].columns)"
      ],
      "id": "8c1e05a9-ba86-4067-9a84-4f9cfcde0bcd"
    },
    {
      "cell_type": "code",
      "execution_count": 21,
      "metadata": {},
      "outputs": [],
      "source": [
        "dict1 = makedict(y)\n",
        "loader = Loader(dict1)\n",
        "dataset = loader.get_dataset(lags=4)\n",
        "train_dataset, test_dataset = temporal_signal_split(dataset, train_ratio=0.8)"
      ],
      "id": "a40c0cea-cff9-4fc8-a4b9-004331d0c74d"
    },
    {
      "cell_type": "code",
      "execution_count": 22,
      "metadata": {},
      "outputs": [],
      "source": [
        "np.array(train_dataset.features).shape, np.array(test_dataset.features).shape"
      ],
      "id": "3e6cd817-9151-4217-b6bc-f0a690727b13"
    },
    {
      "cell_type": "code",
      "execution_count": 56,
      "metadata": {},
      "outputs": [],
      "source": [
        "class RecurrentGCN(torch.nn.Module):\n",
        "    def __init__(self, node_features):\n",
        "        super(RecurrentGCN, self).__init__()\n",
        "        self.recurrent = AGCRN(number_of_nodes = 44,\n",
        "                              in_channels = 4, # 노드특징 수\n",
        "                              out_channels = 16,  # 필터 수\n",
        "                              K = 2,\n",
        "                              embedding_dimensions = 20)\n",
        "        self.linear = torch.nn.Linear(16, 1)\n",
        "\n",
        "    def forward(self, x, e, h):\n",
        "        h_0 = self.recurrent(x, e, h)\n",
        "        y = F.relu(h_0)\n",
        "        y = self.linear(y)\n",
        "        return y, h_0"
      ],
      "id": "2cb2d68a-8f78-4400-84b1-325e264d373b"
    },
    {
      "cell_type": "code",
      "execution_count": 42,
      "metadata": {},
      "outputs": [],
      "source": [
        "# model = RecurrentGCN(node_features = 4)\n",
        "\n",
        "# optimizer = torch.optim.Adam(model.parameters(), lr=0.01)\n",
        "\n",
        "# model.train()\n",
        "\n",
        "# e = torch.empty(44, 4) # 노드수 x 임베딩 차원 수\n",
        "\n",
        "# torch.nn.init.xavier_uniform_(e) # 임베딩 벡터 초기화\n",
        "\n",
        "# for epoch in tqdm(range(1)):\n",
        "#     cost = 0\n",
        "#     h = None\n",
        "#     for time, snapshot in enumerate(train_dataset):\n",
        "#         x = snapshot.x.view(1, 44, 4)\n",
        "#         print(x.shape)\n",
        "    #     y_hat, h = model(x, e, h)\n",
        "    #     cost = cost + torch.mean((y_hat-snapshot.y)**2)\n",
        "    # cost = cost / (time+1)\n",
        "    # cost.backward()\n",
        "    # optimizer.step()\n",
        "    # optimizer.zero_grad()"
      ],
      "id": "72565326-0cd7-4579-8655-3ad7b540ae85"
    },
    {
      "cell_type": "code",
      "execution_count": 57,
      "metadata": {},
      "outputs": [],
      "source": [
        "def get_batches(dataset, len_tr, batch_size=256):\n",
        "    num_batches =  len_tr // batch_size + (1 if len_tr % batch_size != 0 else 0)\n",
        "    batches = []\n",
        "    for i in range(num_batches):\n",
        "        start_idx = i * batch_size\n",
        "        end_idx = start_idx + batch_size\n",
        "        batches.append(train_dataset[start_idx:end_idx])\n",
        "    return batches"
      ],
      "id": "9b69ed1d-aa12-41dd-bec1-45cc143710c2"
    },
    {
      "cell_type": "code",
      "execution_count": 58,
      "metadata": {},
      "outputs": [],
      "source": [
        "len_tr = train_dataset.snapshot_count"
      ],
      "id": "78d58e0a-d005-43a3-aa45-fe1a078b015c"
    },
    {
      "cell_type": "code",
      "execution_count": 59,
      "metadata": {},
      "outputs": [],
      "source": [
        "batches = get_batches(train_dataset, len_tr, batch_size=128)"
      ],
      "id": "57059c97-fe7e-4f73-8178-814fec9057d8"
    },
    {
      "cell_type": "code",
      "execution_count": 62,
      "metadata": {},
      "outputs": [
        {
          "output_type": "stream",
          "name": "stderr",
          "text": [
            "  0%|          | 0/5 [00:00<?, ?it/s] 20%|██        | 1/5 [01:37<06:31, 97.91s/it] 40%|████      | 2/5 [03:28<05:16, 105.52s/it] 60%|██████    | 3/5 [05:42<03:56, 118.42s/it] 80%|████████  | 4/5 [07:33<01:55, 115.62s/it]100%|██████████| 5/5 [10:49<00:00, 129.97s/it]"
          ]
        },
        {
          "output_type": "stream",
          "name": "stdout",
          "text": [
            "epoch=0  loss=1.244255702285205747827       4       \n",
            "epoch=1  loss=0.4966041831409239243774  94  \n",
            "epoch=2  loss=0.2456967795596403263489  68  \n",
            "epoch=3  loss=0.1712142917163231950659  23  \n",
            "epoch=4  loss=0.1428124369943843393677      5   "
          ]
        }
      ],
      "source": [
        "model = RecurrentGCN(node_features = 4)\n",
        "\n",
        "optimizer = torch.optim.Adam(model.parameters(), lr=0.01)\n",
        "\n",
        "model.train()\n",
        "\n",
        "e = torch.empty(44, 20) # 노드수 x 임베딩 차원 수\n",
        "\n",
        "torch.nn.init.xavier_uniform_(e) # 임베딩 벡터 초기화\n",
        "\n",
        "for epoch in tqdm(range(5)):\n",
        "    losses_batch = []\n",
        "    for b, batch in enumerate(batches):\n",
        "        loss = 0\n",
        "        h = None\n",
        "        for time, snapshot in enumerate(batch):\n",
        "            x = snapshot.x.view(1, 44, 4)\n",
        "            y_hat, h = model(x, e, h)\n",
        "            loss = loss + torch.mean((y_hat.reshape(-1)-snapshot.y.reshape(-1))**2)\n",
        "            print(f'\\rbatch={b}\\t t={time+1}\\t loss={loss/(time+1)}\\t', end='', flush=True)\n",
        "        loss = loss / (time+1)\n",
        "        loss.backward()\n",
        "        optimizer.step()\n",
        "        optimizer.zero_grad()\n",
        "        losses_batch.append(loss.item())\n",
        "    print(f'\\repoch={epoch}\\t loss={np.mean(losses_batch)}\\n', end='', flush=True)"
      ],
      "id": "f245561e-ccb0-4b36-892a-9d445936e50a"
    },
    {
      "cell_type": "code",
      "execution_count": 63,
      "metadata": {},
      "outputs": [
        {
          "output_type": "stream",
          "name": "stdout",
          "text": [
            "MSE: 0.2140"
          ]
        }
      ],
      "source": [
        "model.eval()\n",
        "cost = 0\n",
        "for time, snapshot in enumerate(test_dataset):\n",
        "    x = snapshot.x.view(1, 44, 4)\n",
        "    y_hat, h = model(x, e, h)\n",
        "    cost = cost + torch.mean((y_hat-snapshot.y)**2)\n",
        "cost = cost / (time+1)\n",
        "cost = cost.item()\n",
        "print(\"MSE: {:.4f}\".format(cost))"
      ],
      "id": "cabce758-50e1-44c4-8f50-4288db469c60"
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "------------------------------------------------------------------------"
      ],
      "id": "ecb0a476-058a-4d32-8313-7bda87d14eaa"
    },
    {
      "cell_type": "code",
      "execution_count": 64,
      "metadata": {},
      "outputs": [],
      "source": [
        "import torch\n",
        "import torch.nn.functional as F\n",
        "from torch_geometric_temporal.nn.recurrent import A3TGCN\n",
        "\n",
        "from torch_geometric_temporal.dataset import ChickenpoxDatasetLoader\n",
        "from torch_geometric_temporal.signal import temporal_signal_split"
      ],
      "id": "629cee63-ab22-4b9f-be3d-11bf57177490"
    },
    {
      "cell_type": "code",
      "execution_count": 65,
      "metadata": {},
      "outputs": [],
      "source": [
        "loader = ChickenpoxDatasetLoader()\n",
        "\n",
        "dataset = loader.get_dataset()\n",
        "\n",
        "train_dataset, test_dataset = temporal_signal_split(dataset, train_ratio=0.2)"
      ],
      "id": "9c5fa1f1-7ca6-4866-a8e2-4739c95fd08b"
    },
    {
      "cell_type": "code",
      "execution_count": 88,
      "metadata": {},
      "outputs": [],
      "source": [
        "next(iter(dataset))"
      ],
      "id": "90df3d0e-e8c3-43e2-a60f-67bf46d5f271"
    },
    {
      "cell_type": "code",
      "execution_count": 66,
      "metadata": {},
      "outputs": [],
      "source": [
        "class RecurrentGCN(torch.nn.Module):\n",
        "    def __init__(self, node_features, periods):\n",
        "        super(RecurrentGCN, self).__init__()\n",
        "        self.recurrent = A3TGCN(node_features, 32, periods)\n",
        "        self.linear = torch.nn.Linear(32, 1)\n",
        "\n",
        "    def forward(self, x, edge_index, edge_weight):\n",
        "        h = self.recurrent(x.view(x.shape[0], 1, x.shape[1]), edge_index, edge_weight)\n",
        "        h = F.relu(h)\n",
        "        h = self.linear(h)\n",
        "        return h"
      ],
      "id": "e65bb2e3-ef87-426b-ae8e-945237d1589f"
    },
    {
      "cell_type": "code",
      "execution_count": 67,
      "metadata": {},
      "outputs": [
        {
          "output_type": "stream",
          "name": "stderr",
          "text": [
            "100%|██████████| 10/10 [03:40<00:00, 22.07s/it]"
          ]
        }
      ],
      "source": [
        "model = RecurrentGCN(node_features = 1, periods = 4)\n",
        "\n",
        "optimizer = torch.optim.Adam(model.parameters(), lr=0.01)\n",
        "\n",
        "model.train()\n",
        "\n",
        "for epoch in tqdm(range(10)):\n",
        "    cost = 0\n",
        "    for time, snapshot in enumerate(train_dataset):\n",
        "        y_hat = model(snapshot.x, snapshot.edge_index, snapshot.edge_attr)\n",
        "        cost = cost + torch.mean((y_hat-snapshot.y)**2)\n",
        "    cost = cost / (time+1)\n",
        "    cost.backward()\n",
        "    optimizer.step()\n",
        "    optimizer.zero_grad()"
      ],
      "id": "b547ac6c-be3f-4cf8-b61e-a165e95dffe0"
    },
    {
      "cell_type": "code",
      "execution_count": 68,
      "metadata": {},
      "outputs": [
        {
          "output_type": "stream",
          "name": "stdout",
          "text": [
            "MSE: 1.0110"
          ]
        }
      ],
      "source": [
        "model.eval()\n",
        "cost = 0\n",
        "for time, snapshot in enumerate(test_dataset):\n",
        "    y_hat = model(snapshot.x, snapshot.edge_index, snapshot.edge_attr)\n",
        "    cost = cost + torch.mean((y_hat-snapshot.y)**2)\n",
        "cost = cost / (time+1)\n",
        "cost = cost.item()\n",
        "print(\"MSE: {:.4f}\".format(cost))"
      ],
      "id": "76bba499-ba9b-40ac-826f-66eff0bd1527"
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "`-` solar data"
      ],
      "id": "16c6c020-d7a6-4f1c-b13f-bc0c0e18ec03"
    },
    {
      "cell_type": "code",
      "execution_count": 69,
      "metadata": {},
      "outputs": [],
      "source": [
        "# read dataframe \n",
        "df = pd.read_csv('../SOLAR/data_eng_230710.csv')\n",
        "\n",
        "# make y, y_upper, y_period, time, regions \n",
        "y = df.loc[:,'Bukchoncheon':'Gyeongju-si'].to_numpy()\n",
        "yU = df.loc[:,'Bukchoncheon_Upper':'Gyeongju-si_Upper'].to_numpy()\n",
        "yP = np.divide(y, yU+1e-10)\n",
        "\n",
        "t = df.loc[:,'date']\n",
        "regions = list(df.loc[:, 'Bukchoncheon':'Gyeongju-si'].columns)"
      ],
      "id": "662a505d-13d3-49a9-a061-7c50ec6e1911"
    },
    {
      "cell_type": "code",
      "execution_count": 82,
      "metadata": {},
      "outputs": [],
      "source": [
        "dict1 = makedict(y)\n",
        "loader = Loader(dict1)\n",
        "dataset = loader.get_dataset(lags=1)\n",
        "train_dataset, test_dataset = temporal_signal_split(dataset, train_ratio=0.8)"
      ],
      "id": "e6494d5a-80a4-4601-8565-51c7e9dc6a50"
    },
    {
      "cell_type": "code",
      "execution_count": 83,
      "metadata": {},
      "outputs": [],
      "source": [
        "np.array(train_dataset.features).shape, np.array(test_dataset.features).shape"
      ],
      "id": "4783d24b-c7dc-4f2f-b215-b225887b1c8f"
    },
    {
      "cell_type": "code",
      "execution_count": 84,
      "metadata": {},
      "outputs": [],
      "source": [
        "len_tr = train_dataset.snapshot_count"
      ],
      "id": "a743a441-6060-409b-bc7e-653cef14b773"
    },
    {
      "cell_type": "code",
      "execution_count": 85,
      "metadata": {},
      "outputs": [],
      "source": [
        "batches = get_batches(train_dataset, len_tr, batch_size=128)"
      ],
      "id": "e4bb062f-c91d-4ce8-927f-99d022276b7d"
    },
    {
      "cell_type": "code",
      "execution_count": 86,
      "metadata": {},
      "outputs": [],
      "source": [
        "class RecurrentGCN(torch.nn.Module):\n",
        "    def __init__(self, node_features, periods):\n",
        "        super(RecurrentGCN, self).__init__()\n",
        "        self.recurrent = A3TGCN(node_features, 32, periods)\n",
        "        self.linear = torch.nn.Linear(32, 1)\n",
        "\n",
        "    def forward(self, x, edge_index, edge_weight):\n",
        "        h = self.recurrent(x.view(x.shape[0], 1, x.shape[1]), edge_index, edge_weight)\n",
        "        h = F.relu(h)\n",
        "        h = self.linear(h)\n",
        "        return h"
      ],
      "id": "7a6b4e9e-2fb5-46af-a39c-5f51c1a68d81"
    },
    {
      "cell_type": "code",
      "execution_count": 87,
      "metadata": {},
      "outputs": [
        {
          "output_type": "stream",
          "name": "stderr",
          "text": [
            "  0%|          | 0/5 [00:00<?, ?it/s]"
          ]
        }
      ],
      "source": [
        "model = RecurrentGCN(node_features = 1, periods = 4)\n",
        "\n",
        "optimizer = torch.optim.Adam(model.parameters(), lr=0.01)\n",
        "\n",
        "model.train()\n",
        "\n",
        "for epoch in tqdm(range(5)):\n",
        "    losses_batch = []\n",
        "    for b, batch in enumerate(batches):\n",
        "        loss = 0\n",
        "        h = None\n",
        "        for time, snapshot in enumerate(batch):\n",
        "            y_hat = model(snapshot.x, snapshot.edge_index, snapshot.edge_attr)\n",
        "            loss = loss + torch.mean((y_hat.reshape(-1)-snapshot.y.reshape(-1))**2)\n",
        "            print(f'\\rbatch={b}\\t t={time+1}\\t loss={loss/(time+1)}\\t', end='', flush=True)\n",
        "        loss = loss / (time+1)\n",
        "        loss.backward()\n",
        "        optimizer.step()\n",
        "        optimizer.zero_grad()\n",
        "        losses_batch.append(loss.item())\n",
        "    print(f'\\repoch={epoch}\\t loss={np.mean(losses_batch)}\\n', end='', flush=True)"
      ],
      "id": "e15da001-22fe-499a-95e6-59491c31f9f9"
    }
  ],
  "nbformat": 4,
  "nbformat_minor": 5,
  "metadata": {
    "kernelspec": {
      "name": "python3",
      "display_name": "Python 3",
      "language": "python"
    },
    "language_info": {
      "name": "python",
      "codemirror_mode": {
        "name": "ipython",
        "version": "3"
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.8.16"
    }
  }
}